#include "pch.h"
#ifdef bbcDEBUG_TRACE
#include "bbcDebugTrace.h"
#include "bbcComputer.h"

enum {
	chunk_size=500000,
};

bbcDebugTrace::bbcDebugTrace():
num_entries_(0)
{
	this->AddChunk();
}

bbcDebugTrace::~bbcDebugTrace() {
}

void bbcDebugTrace::AddChunk() {
	chunks_.push_back(Chunk());
	chunks_.back().resize(chunk_size);
	chunk_index_=0;
}

bbcDebugTrace::Ent *bbcDebugTrace::NewEnt(Type type) {
	if(chunk_index_==chunk_size) {
		this->AddChunk();
	}
	Ent *ent=&chunks_.back()[chunk_index_];
	++chunk_index_;
	ent->type=type;
	ent->cycles=bbcComputer::cpucycles;
	ent->next_stop=bbcComputer::next_stop;
	++num_entries_;
	return ent;
}

#define IRQ_FLAG(X) (irq->irqflags&(X)?#X:"")
	
bool bbcDebugTrace::WriteTrace(FILE *out,ProgressFn progress_fn,void *context) const {
	int hundredth=num_entries_/100;
	unsigned n=0;

	for(Chunks::const_iterator chunk=chunks_.begin();chunk!=chunks_.end();++chunk) {
		for(unsigned i=0;n<num_entries_&&i<chunk->size();++i,++n) {
			const Ent *ent=&(*chunk)[i];
			const bbcDebugTrace::Irq *irq=&ent->irq;
			const bbcDebugTrace::Instr *ins=&ent->instr;
			
			if(progress_fn&&n%hundredth==0) {
				if(!(*progress_fn)(n/hundredth,context)) {
					//Cancelled
					return false;
				}
			}
			int left_margin=fprintf(out,"%-9d (@%-9d) ",ent->cycles,ent->next_stop);
			switch(ent->type) {
			default:
				BASSERT(false);
				fprintf(out,"unknown entry type %d (shouldn't ever see this)\n",
					ent->type);
				break;
			case bbcDebugTrace::ENT_IRQ:
				fprintf(out,"flags 0x%X: %s %s %s %s\n",
					irq->irqflags,IRQ_FLAG(IRQ_SYSVIA),IRQ_FLAG(IRQ_USERVIA),IRQ_FLAG(IRQ_ACIA),IRQ_FLAG(IRQ_NMI));
				this->WriteViaState(out,left_margin,"SysVIA",&irq->sys);
				this->WriteViaState(out,left_margin,"UsrVIA",&irq->usr);
				fprintf(out,"%-*s1770  : Status=&%02X\n",
					left_margin,"",irq->fdc_status);
				break;
			case bbcDebugTrace::ENT_INSTR:
				{
					char opcode_buf[10],operand_buf[15];
					int len=bbcComputer::DisassembleBytes(ins->bytes,ins->pc,opcode_buf,
						operand_buf);
					const char *nullbytes="  ";
					char bytes[3][3];
					static const char hex_digits[]="0123456789ABCDEF";
					for(int i=0;i<3;++i) {
						bytes[i][0]=hex_digits[ins->bytes[i]>>4];
						bytes[i][1]=hex_digits[ins->bytes[i]&0xf];
						bytes[i][2]=0;
					}
					
					fprintf(out,"%04X: %-5s %-10s %s %s %s -> A:%02X X:%02X Y:%02X S:%02X P:%c%c%c%c%c%c%c%c\n",
						ins->pc.w,opcode_buf,operand_buf,
						len>0?bytes[0]:nullbytes,
						len>1?bytes[1]:nullbytes,
						len>2?bytes[2]:nullbytes,
						ins->a,ins->x,ins->y,ins->s,
						ins->p&t65::N_MASK?'N':'-',
						ins->p&t65::V_MASK?'V':'-',
						ins->p&t65::U_MASK?'-':'-',
						ins->p&t65::B_MASK?'B':'-',
						ins->p&t65::D_MASK?'D':'-',
						ins->p&t65::I_MASK?'I':'-',
						ins->p&t65::Z_MASK?'Z':'-',
						ins->p&t65::C_MASK?'C':'-');
				}
				break;
			}
		}
	}
	printf(__FUNCTION__ ": wrote %u lines\n",n);
	return true;
}

static const char *const via_irq_flag_names[]={
	"T1","T2","CB1","CB2","SR","CA1","CA2",
};

static void WriteViaIrqFlags(FILE *out,int indent,const char *reg,t65::byte flags) {
	fprintf(out,"%-*s: %s:",indent,"",reg);
	for(unsigned i=0;i<7;++i) {
		fprintf(out,"%-4s",(flags<<i)&0x40?via_irq_flag_names[i]:"");
	}
	fprintf(out,"\n");
}

void bbcDebugTrace::WriteViaState(FILE *out,int indent,const char *via_name,const bbcVIA::TraceState *state) const {
	fprintf(out,"%-*s%s: IER=&%02X IFR=&%02X ACR=&%02X T1L=&%02X%02X T1C=%d T2C=%d\n",
		indent,"",via_name,state->ier,state->ifr,state->acr,state->t1l.h,state->t1l.l,state->t1c,state->t2c);
	int irq_indent=indent+strlen(via_name);
	WriteViaIrqFlags(out,irq_indent,"IER",state->ier);
	WriteViaIrqFlags(out,irq_indent,"IFR",state->ifr);
}

#endif//bbcDEBUG_TRACE
