#include "pch.h"
#include "bbcModel.h"
#include "bbcComputer.h"
#include "bbc1770.h"
#include "bbcSystemVIA.h"
#include "bbcUserVIA.h"
#include <stdarg.h>

#ifdef bbcDEBUG_TRACE_EVENTS
//////////////////////////////////////////////////////////////////////////
//
static const char hex_digits[]="0123456789ABCDEF";

const bbcModelCpuEventHandler bbcModelCpuEventHandler::handler;

bbcModelCpuEventHandler::bbcModelCpuEventHandler():
bbcEventHandler(sizeof(Data),"CPU ")
{
}

void bbcModelCpuEventHandler::FillEventData(void *event) const {
	//Doesn't do anything
}

void bbcModelCpuEventHandler::AppendEventStr(const bbcEventInfo *info,const void *event,int indent,std::vector<char> *output) const {
	static const char *nullbytes="  ";

	const Data *ev=static_cast<const Data *>(event);
	char linebuf[80];
	char opcode_buf[10],operand_buf[15];
	int len=bbcComputer::DisassembleBytes(ev->bytes,ev->pc,opcode_buf,operand_buf);

	char bytes[3][3];
	for(int i=0;i<3;++i) {
		bytes[i][0]=hex_digits[ev->bytes[i]>>4];
		bytes[i][1]=hex_digits[ev->bytes[i]&0xf];
		bytes[i][2]=0;
	}

	//0123456789012345678901234567890123456789012345678901234567890123456789
	//1234: 12345 1234567890 xx xx xx -> A:xx X:xx Y:xx S:xx P:12345678
	int n=_snprintf(linebuf,sizeof linebuf-1,"%04X: %-5s %-10s %s %s %s -> A:%02X X:%02X Y:%02X S:%02X P:%c%c%c%c%c%c%c%c\n",
		ev->pc.w,opcode_buf,operand_buf,
		len>0?bytes[0]:nullbytes,
		len>1?bytes[1]:nullbytes,
		len>2?bytes[2]:nullbytes,
		ev->a,ev->x,ev->y,ev->s,
		ev->p&t65::N_MASK?'N':'-',
		ev->p&t65::V_MASK?'V':'-',
		ev->p&t65::U_MASK?'-':'-',
		ev->p&t65::B_MASK?'B':'-',
		ev->p&t65::D_MASK?'D':'-',
		ev->p&t65::I_MASK?'I':'-',
		ev->p&t65::Z_MASK?'Z':'-',
		ev->p&t65::C_MASK?'C':'-');

	output->insert(output->end(),linebuf,linebuf+n);
}

//////////////////////////////////////////////////////////////////////////
//
const bbcModelIrqEventHandler bbcModelIrqEventHandler::handler;

bbcModelIrqEventHandler::bbcModelIrqEventHandler():
bbcEventHandler(sizeof(Data),"IRQ ")
{
}

//grab irq data from wherever.
//(this is all getting disturbingly cosy.)
void bbcModelIrqEventHandler::FillEventData(void *event) const {
	Data *ev=static_cast<Data *>(event);

	ev->fdc_status=bbc1770::Status();
	ev->irqflags=bbcComputer::irq_flags_;
	bbc_system_via.FillTraceState(&ev->sys);
	bbc_user_via.FillTraceState(&ev->usr);
}

//static const char *const via_irq_flag_names[]={
//	"T1","T2","CB1","CB2","SR","CA1","CA2",
//};
//
//static void WriteViaIrqFlags(FILE *out,int indent,const char *reg,t65::byte flags) {
//	fprintf(out,"%-*s: %s:",indent,"",reg);
//	for(unsigned i=0;i<7;++i) {
//		fprintf(out,"%-4s",(flags<<i)&0x40?via_irq_flag_names[i]:"");
//	}
//	fprintf(out,"\n");
//}
//
//void bbcDebugTrace::WriteViaState(FILE *out,int indent,const char *via_name,const Irq::Via *state) const {
//	fprintf(out,"%-*s%s: IER=&%02X IFR=&%02X ACR=&%02X T1L=&%02X%02X T1C=%d T2C=%d\n",
//		indent,"",via_name,state->ier,state->ifr,state->acr,state->t1l.h,state->t1l.l,state->t1c,state->t2c);
//	int irq_indent=indent+strlen(via_name);
//	WriteViaIrqFlags(out,irq_indent,"IER",state->ier);
//	WriteViaIrqFlags(out,irq_indent,"IFR",state->ifr);
//}
//

#define IRQ_FLAG(F) (ev->irqflags&(F)?" " #F:"")

static void AppendFmt(std::vector<char> *output,const char *fmt,...) {
	va_list v;
	static const unsigned max_size=150;

	unsigned old_size=output->size();
	output->resize(old_size+max_size);//make space
	va_start(v,fmt);
	int n=_vsnprintf(&(*output)[old_size],max_size,fmt,v);
	va_end(v);

	//chop off the excess.
	output->resize(old_size+n);
}

void bbcModelIrqEventHandler::AppendViaFlagsState(t65::byte val,const char *name,int indent,std::vector<char> *output) const {
	static const char *const via_irq_flag_names[]={"T1","T2","CB1","CB2","SR","CA1","CA2",};
	static const char *blank="   ";//must be same length as longest flag name

	output->insert(output->end(),indent,' ');
	output->insert(output->end(),name,name+3);
	for(unsigned i=0;i<7;++i) {
		t65::byte mask=0x40>>i;
		const char *src=val&mask?via_irq_flag_names[i]:blank;

		output->push_back(' ');
		for(unsigned j=0;via_irq_flag_names[i][j];++j) {
			output->push_back(src[j]);
		}
	}
	output->push_back('\n');
}

void bbcModelIrqEventHandler::AppendViaState(const bbcVIA::TraceState &state,const char *name,int indent,std::vector<char> *output) const {
	
	//line 1, basic state
	output->insert(output->end(),indent,' ');
	AppendFmt(output,"%s: IER=&%02X IFR=&%02X ACR=&%02X T1L=&%02X%02X T1C=%d T2C=%d\n",
		name,state.ier,state.ifr,state.acr,state.t1l.h,state.t1l.l,state.t1c,state.t2c);

	//line 2, IER state
	indent+=strlen(name)+2;//+2 for ": "
	this->AppendViaFlagsState(state.ier,"IER",indent,output);

	//line 3, IFR state
	this->AppendViaFlagsState(state.ifr,"IFR",indent,output);
}

void bbcModelIrqEventHandler::AppendEventStr(const bbcEventInfo *info,const void *event,int indent,std::vector<char> *output) const {
	const Data *ev=static_cast<const Data *>(event);
	
	AppendFmt(output,"flags 0x%X: %s%s%s%s\n",ev->irqflags,IRQ_FLAG(IRQ_SYSVIA),IRQ_FLAG(IRQ_USERVIA),IRQ_FLAG(IRQ_ACIA),IRQ_FLAG(IRQ_NMI));
	this->AppendViaState(ev->sys,"SysVIA",indent,output);
	this->AppendViaState(ev->usr,"UsrVIA",indent,output);
	AppendFmt(output,"%-*s1770  : Status=&%02X\n",indent,"",ev->fdc_status);

//	fprintf(out,"flags 0x%X: %s %s %s %s\n",
//		irq->irqflags,IRQ_FLAG(IRQ_SYSVIA),IRQ_FLAG(IRQ_USERVIA),IRQ_FLAG(IRQ_ACIA),IRQ_FLAG(IRQ_NMI));
//	this->WriteViaState(out,left_margin,"SysVIA",&irq->sys);
//	this->WriteViaState(out,left_margin,"UsrVIA",&irq->usr);
//	fprintf(out,"%-*s1770  : Status=&%02X\n",
//		left_margin,"",irq->fdc_status);

}
#endif

//////////////////////////////////////////////////////////////////////////
//
bbcModel::bbcModel(unsigned flags_in):
flags(flags_in)
{
}

bbcModel::~bbcModel() {
}

