#include "pch.h"
#include "mbConfig.h"
#include "mbModel.h"
#include "mbKeymap.h"
#include "Host.h"
#include "mbMisc.h"
#include "mbKeys.h"
#include "mbDiscInterface.h"
#include <algorithm>
#include <iostream>
#include <sstream>
#include <bbcVideo.h>
#include <bbcSound.h>

//////////////////////////////////////////////////////////////////////////
static void LoadRect(mbWxConfig *cfg,const wxString &dir,wxRect *rect) {
	wxASSERT(rect);

	if(!cfg->Read(dir+"/x",&rect->x)||!cfg->Read(dir+"/y",&rect->y)||
		!cfg->Read(dir+"/width",&rect->width)||!cfg->Read(dir+"/height",&rect->height))
	{
		*rect=wxRect(-1,-1,-1,-1);
	}
}

static void SaveRect(mbWxConfig *cfg,const wxString &dir,const wxRect &rect) {
	cfg->Write(dir+"/x",rect.x);
	cfg->Write(dir+"/y",rect.y);
	cfg->Write(dir+"/width",rect.width);
	cfg->Write(dir+"/height",rect.height);
}

template<class T>
static void ClampValue(const wxString &name,T *value,const T &mini,const T &maxi) {
	T old_value=*value;
	if(*value<mini) {
		*value=mini;
	} else if(*value>maxi&&maxi>mini) {
		*value=maxi;
	}
	if(*value!=old_value) {
		std::ostringstream old_val,new_val;
		old_val<<old_value<<std::ends;
		new_val<<*value<<std::ends;
		wxLogWarning("%s: invalid value %s, set to %s",name.c_str(),old_val.str().c_str(),
			new_val.str().c_str());
	}
}

//////////////////////////////////////////////////////////////////////////
//
mbRomsConfig::Rom::Rom():
ram(false)
{
}

mbRomsConfig::mbRomsConfig():
dlg_rect(-1,-1,-1,-1),
fs_slot(-1)
{
}

void mbRomsConfig::LoadRom(mbWxConfig *cfg,const wxString &key,Rom *rom) {
	rom->ram=false;
	wxString value;
	if(cfg->Read(key,&value)) {
		size_t comma=value.find_last_of(',');
		if(comma!=wxString::npos) {
			if(value.substr(comma).Lower()==",ram") {
				rom->ram=true;
			}
			value=value.substr(0,comma);
		}
		rom->filename=value;
	}
}

void mbRomsConfig::SaveRom(mbWxConfig *cfg,const wxString &key,const Rom *rom) const {
	wxString value=mbRelativeFileName(rom->filename,".");
	if(rom->ram) {
		value+=",ram";
	}
	cfg->Write(key,value);
}

//helper function for loading and saving roms.
//returns pointer to member of mbRomsConfig -- the string that is the OS name for this model. returns NULL if we've
//already seen that OS. index is the index into models of the model we're currently looking at.
static wxString mbRomsConfig::*UniqueModelOsName(const std::vector<const mbModel *> models,unsigned index) {
	unsigned j;

	for(j=0;j<index;++j) {
		if(models[index]->os_name==models[j]->os_name) {
			return 0;
		}
	}

	return models[index]->os_name;
}

void mbRomsConfig::Load(mbWxConfig *cfg) {
	unsigned i;
//	cfg->Read("/Roms/Os",&this->os_filename);

	std::vector<const mbModel *> models;
	mbModelGetAll(&models);
	for(i=0;i<models.size();++i) {
		wxString mbRomsConfig::*os_name=UniqueModelOsName(models,i);
		if(os_name) {
			cfg->Read(wxString::Format("/Roms/%s_os",models[i]->short_name),&(this->*os_name));
		}
		//cfg->Read(wxString::Format("/Roms/%s_os",models[i]->short_name),&(this->*models[i]->os_name));
		//Rom *roms[16];
		//models[i]->GetRomsPtrs(*this,roms);
		for(int j=0;j<16;++j) {
			wxString key;

			if((models[i]->b_rom_mask)&(1<<j)) {
				//B style ROM
				key=wxString::Format("/Roms/%d",j);
			} else {
				//M128 style ROM
				key=wxString::Format("/Roms/%s_%d",models[i]->short_name,j);
			}

			Rom *rom=models[i]->CfgRomPtr(*this,j);
			this->LoadRom(cfg,key,rom);
		}
	}

	LoadRect(cfg,"/Roms/DialogRect",&this->dlg_rect);
	cfg->Read("/Roms/FsSlot",&this->fs_slot,-1);
	if(this->fs_slot>15) {
		this->fs_slot=-1;
	}

	this->fs_roms.clear();
	std::vector<const mbDiscInterface *> ifs;
	mbDiscInterfaceGetAll(&ifs);
	for(i=0;i<ifs.size();++i) {
		wxString rom;

		cfg->Read(wxString::Format("/Roms/fs_rom_%s",ifs[i]->short_name),&rom,"");
		if(!rom.empty()) {
			this->fs_roms[ifs[i]->short_name]=rom;
		}
	}
}

void mbRomsConfig::Save(mbWxConfig *cfg) const {
	std::vector<const mbModel *> models;
	mbModelGetAll(&models);
	for(unsigned i=0;i<models.size();++i) {
		const wxString mbRomsConfig::*os_name=UniqueModelOsName(models,i);
		if(os_name) {
			cfg->Write(wxString::Format("/Roms/%s_os",models[i]->short_name),mbRelativeFileName(this->*os_name,"."));
		}

		//const Rom *roms[16];
		//models[i]->GetRomsPtrs(*this,roms);
		for(int j=0;j<16;++j) {
			wxString key;

			if((models[i]->b_rom_mask)&(1<<j)) {
				//B style ROM
				key=wxString::Format("/Roms/%d",j);
			} else {
				//M128 style ROM
				key=wxString::Format("/Roms/%s_%d",models[i]->short_name,j);
			}
			
			const Rom *rom=models[i]->CfgRomPtr(*this,j);
			this->SaveRom(cfg,key,rom);
		}
	}

	SaveRect(cfg,"/Roms/DialogRect",this->dlg_rect);
	cfg->Write("/Roms/FsSlot",this->fs_slot);

	for(std::map<wxString,wxString>::const_iterator it=this->fs_roms.begin();it!=this->fs_roms.end();++it) {
		if(!it->second.empty()) {
			cfg->Write("/Roms/fs_rom_"+it->first,mbRelativeFileName(it->second,"."));
		}
	}
}

//////////////////////////////////////////////////////////////////////////
//
mbSoundConfig::mbSoundConfig():
enabled(false),
hz(44100),
bits(2),
stereo(false),
len_frames(4),
dlg_rect(-1,-1,-1,-1),
volume(100)
{
	for(unsigned i=0;i<4;++i) {
		this->is_stereo_left[i]=true;
		this->is_stereo_right[i]=true;
	}
}

void mbSoundConfig::Load(mbWxConfig *cfg) {
	cfg->Read("/Sound/Enabled",&this->enabled);
	cfg->Read("/Sound/Hz",&this->hz);
	cfg->Read("/Sound/Bits",&this->bits);
	cfg->Read("/Sound/Stereo",&this->stereo);
	cfg->Read("/Sound/LenFrames",&this->len_frames);
	for(unsigned i=0;i<4;++i) {
		cfg->Read(wxString::Format("/Sound/Stereo/%uLeft",i),&this->is_stereo_left[i]);
		cfg->Read(wxString::Format("/Sound/Stereo/%uRight",i),&this->is_stereo_right[i]);
	}
	LoadRect(cfg,"/Sound/DialogRect",&this->dlg_rect);
	cfg->Read("/Sound/Volume",&this->volume);

	ClampValue("Sound/LenFrames",&this->len_frames,3,int(bbcSound::max_len_frames));
}

void mbSoundConfig::Save(mbWxConfig *cfg)  const{
	cfg->Write("/Sound/Enabled",this->enabled);
	cfg->Write("/Sound/Hz",this->hz);
	cfg->Write("/Sound/Bits",this->bits);
	cfg->Write("/Sound/Stereo",this->stereo);
	cfg->Write("/Sound/LenFrames",this->len_frames);
	for(unsigned i=0;i<4;++i) {
		cfg->Write(wxString::Format("/Sound/Stereo/%uLeft",i),this->is_stereo_left[i]);
		cfg->Write(wxString::Format("/Sound/Stereo/%uRight",i),this->is_stereo_right[i]);
	}
	SaveRect(cfg,"/Sound/DialogRect",this->dlg_rect);
	cfg->Write("/Sound/Volume",this->volume);
}

//////////////////////////////////////////////////////////////////////////
//
mbKeyboardConfig::mbKeyboardConfig():
dlg_rect(-1,-1,-1,-1)
{
	std::fill(this->keylinks,this->keylinks+8,false);
}

void mbKeyboardConfig::Load(mbWxConfig *cfg) {
	//Keyboard links
	wxString binlinks;
	if(cfg->Read("/Keyboard/Keylinks",&binlinks)) {
		if(binlinks.length()!=8) {
			wxLogWarning("Bad keylinks settings; setting to defaults.\n");
			for(unsigned i=0;i<8;++i) {
				this->keylinks[i]=false;
			}
		} else {
			for(unsigned i=0;i<8;++i) {
				if(binlinks[i]!='0'&&binlinks[i]!='1') {
					wxLogWarning("Bad character '%c' in keylinks.",binlinks[i]);
				} else {
					this->keylinks[i]=binlinks[i]=='1';
				}
				wxLogVerbose("link %u (bit %u, 0x%X) is %s\n",1+i,7-i,1<<(7-i),
					this->keylinks[i]?"set":"clear");
			}
		}
	}

	wxString old_path=cfg->GetPath();
	cfg->SetPath("/Keyboard");
	wxString group_name;
	long group_index;
	if(cfg->GetFirstGroup(group_name,group_index)) {
		std::vector<wxString> keymap_names;
		do {
			if(group_name!="DialogRect") {
				wxLogDebug(__FUNCTION__ ": group \"%s\"",group_name.c_str());
				keymap_names.push_back(group_name);
			}
		} while(cfg->GetNextGroup(group_name,group_index));

		for(unsigned i=0;i<keymap_names.size();++i) {
			wxLogVerbose("Reading keymap %s...\n",keymap_names[i].c_str());
			mbKeymap new_keymap(keymap_names[i]);
			cfg->SetPath("/Keyboard/"+keymap_names[i]);
			
			std::vector<wxString> key_names;
			wxString key_name;
			long key_index;
			
			if(cfg->GetFirstEntry(key_name,key_index)) {
				do {
					key_names.push_back(key_name);
				} while(cfg->GetNextEntry(key_name,key_index));
			}

			for(unsigned j=0;j<key_names.size();++j) {
				wxString host_keys;

				cfg->Read(key_names[j],&host_keys);

				int bbc_code=mbCodeFromKeyName(key_names[j].c_str());
				if(bbc_code==bbcKEY_NONE) {
					wxLogWarning("Invalid key %s",key_names[j].c_str());
				} else {
					std::vector<wxString> keys;
					mbTokenizeString(host_keys,",",keys,true);

//					wxLogDebug(__FUNCTION__ ": Keys for %s:",key_names[j].c_str());
					for(unsigned k=0;k<keys.size();++k) {
						size_t n=keys[k].find_first_of(":");
						mbKeymap::Key mapping;
						if(n==wxString::npos) {
							mapping.key=keys[k];
//							wxLogDebug(__FUNCTION__ ":     %s",mapping.key.c_str());
						} else {
							mapping.device=keys[k].substr(0,n);
							mapping.key=keys[k].substr(n+1);
//							wxLogDebug(__FUNCTION__ ":     %s:%s",mapping.device.c_str(),mapping.key.c_str());
						}
						new_keymap.keymap[bbc_code].push_back(mapping);
					}
				}
				//new_keymap.SetKey(key_names[j],host_keys);
			}
			this->keymaps.push_back(new_keymap);
		}
	}
	cfg->SetPath(old_path);

	if(this->keymaps.empty()) {
		//this->keymaps.push_back(mbKeymap::DefaultKeymap());
		wxLogWarning("No keymaps found.");// -- adding default UK QWERTY\n");
	}

	cfg->Read("/Keyboard/SelectedKeymap",&this->selected_keymap);
	//this->selected_keymap=cfg->Read("/Keyboard/SelectedKeymap");
	LoadRect(cfg,"/Keyboard/DialogRect",&this->dlg_rect);
}

void mbKeyboardConfig::Save(mbWxConfig *cfg) const {
	wxString binlinks;
	for(unsigned i=0;i<8;++i) {
		binlinks+=this->keylinks[i]?"1":"0";
	}
	cfg->Write("/Keyboard/Keylinks",binlinks);

	mbKeymap::C::const_iterator src;
	for(src=this->keymaps.begin();src!=this->keymaps.end();++src) {
		for(int i=0;i<256;++i) {
			const mbKeymap::HostKeys *srckeys=&src->keymap[i];
			mbKeymap::HostKeys::const_iterator srckey;

			wxString mb_key=mbKeyNameFromCode(i);
			if(!srckeys->empty()&&!mb_key.empty()) {
				wxString all;

				for(srckey=srckeys->begin();srckey!=srckeys->end();++srckey) {
					if(!all.empty()) {
						all+=",";
					}

					if(!srckey->device.empty()) {
						all+=srckey->device+":";
					}

					all+=srckey->key;
				}

				if(!all.empty()) {
//					wxLogDebug("/Keyboard/"+src->name+"/"+mb_key+"="+all);
					cfg->Write("/Keyboard/"+src->name+"/"+mb_key,all);
				}
			}
		}
	}

	cfg->Write("/Keyboard/SelectedKeymap",selected_keymap);
	SaveRect(cfg,"/Keyboard/DialogRect",this->dlg_rect);
}

//	for(mbKeymap::C::const_iterator keymap=this->keymaps.begin();keymap!=this->keymaps.end();
//		++keymap)
//	{
//		for(int i=0;i<256;++i) {
//			const mbKeymap::HostKeys *host_keys=&keymap->keymap[i];
//
//			wxString mb_key=mbKeyNameFromCode(i);
//			if(host_keys->empty()||mb_key.empty()) {
//				continue;
//			}
//
//			wxString host_keys_names;
//			for(mbKeymap::HostKeys::const_iterator host_key=host_keys->begin();
//				host_key!=host_keys->end();++host_key)
//			{
//				if(!host_keys_names.empty()) {
//					host_keys_names+=",";
//				}
//
//				if(!host_key->device.empty()) {
//					host_keys_names+=host_key->device+":";
//				}
//
//				host_keys_names+=host_key->key;
//			}
//			if(!host_keys_names.empty()) {
//				wxLogDebug("/Keyboard/"+keymap->name+"/"+mb_key+"="+host_keys_names);
//				cfg->Write("/Keyboard/"+keymap->name+"/"+mb_key,host_keys_names);
//			}
//		}
//	}
//	cfg->Write("/Keyboard/SelectedKeymap",selected_keymap);
//	SaveRect(cfg,"/Keyboard/DialogRect",this->dlg_rect);
//}

//////////////////////////////////////////////////////////////////////////
//

const int mbVideoConfig::max_frame_skip=50;

static mbScanlinesMode ScanlinesModeFromName(const wxString &str) {
	if(str.CmpNoCase("single")==0) {
		return mbSCANLINES_SINGLE;
	} else if(str.CmpNoCase("interlace")==0) {
		return mbSCANLINES_INTERLACE;
	} else if(str.CmpNoCase("double")==0) {
		return mbSCANLINES_DOUBLE;
	}
	wxLogWarning("Unknown scanlines mode \"%s\" -- using single",str.c_str());
	return mbSCANLINES_SINGLE;
}

static const wxString NameFromScanlinesMode(mbScanlinesMode mode) {
	switch(mode) {
	default:
	case mbSCANLINES_SINGLE:
		return "single";
	case mbSCANLINES_INTERLACE:
		return "interlace";
	case mbSCANLINES_DOUBLE:
		return "double";
	}
}

mbVideoConfig::mbVideoConfig():
width(640),
height(272),
frame_skip(0),
cleanup_frequency(100),
full_screen(false),
full_screen_refresh_rate(0),
full_screen_vsync(false),
full_screen_scanlines(mbSCANLINES_SINGLE),
windowed_scanlines(mbSCANLINES_SINGLE),
monitor_type(mbMONITOR_COLOUR),
video_effect(0)
{
}

void mbVideoConfig::Load(mbWxConfig *cfg) {
	wxString scanlines_mode_name;
	int monitor;

	cfg->Read("/Video/Width",&this->width);
	cfg->Read("/Video/Height",&this->height);
	cfg->Read("/Video/FrameSkip",&this->frame_skip);
	cfg->Read("/Video/CleanupFrequency",&this->cleanup_frequency);
	//cfg->Read("/Video/Vertical2x",&this->vertical_2x);
	cfg->Read("/Video/FullScreen",&this->full_screen);
	cfg->Read("/Video/FullScreenRefreshRate",&this->full_screen_refresh_rate);
	cfg->Read("/Video/FullScreenVsync",&this->full_screen_vsync);
	cfg->Read("/Video/MonitorType",&monitor);
	cfg->Read("/Video/VideoEffect",&this->video_effect);

	scanlines_mode_name="";
	cfg->Read("/Video/FullScreenScanlinesMode",&scanlines_mode_name);
	this->full_screen_scanlines=ScanlinesModeFromName(scanlines_mode_name);
	scanlines_mode_name="";
	cfg->Read("/Video/WindowedScanlinesMode",&scanlines_mode_name);
	this->windowed_scanlines=ScanlinesModeFromName(scanlines_mode_name);
	
	ClampValue("Video/Width",&this->width,bbcVideo::min_buffer_width,bbcVideo::max_buffer_width);
	ClampValue("Video/Height",&this->height,bbcVideo::min_buffer_height,bbcVideo::max_buffer_height);
	ClampValue("Video/FrameSkip",&this->frame_skip,1,max_frame_skip);
	ClampValue("Video/CleanupFrequency",&this->cleanup_frequency,1,-1);
	ClampValue("Video/MonitorType",&monitor,0,mbMONITOR_NUM_TYPES-1);

	LoadRect(cfg,"/Video/DialogRect",&this->dlg_rect);
	this->monitor_type=mbMonitorType(monitor);
}

void mbVideoConfig::Save(mbWxConfig *cfg) const {
	cfg->Write("/Video/Width",this->width);
	cfg->Write("/Video/Height",this->height);
	cfg->Write("/Video/FrameSkip",this->frame_skip);
	cfg->Write("/Video/CleanupFrequency",this->cleanup_frequency);
	//cfg->Write("/Video/Vertical2x",this->vertical_2x);
	cfg->Write("/Video/FullScreen",this->full_screen);
	cfg->Write("/Video/FullScreenScanlinesMode",NameFromScanlinesMode(this->full_screen_scanlines));
	cfg->Write("/Video/WindowedScanlinesMode",NameFromScanlinesMode(this->windowed_scanlines));
	cfg->Write("/Video/FullScreenRefreshRate",this->full_screen_refresh_rate);
	cfg->Write("/Video/FullScreenVsync",this->full_screen_vsync);
	cfg->Write("/Video/MonitorType",this->monitor_type);
	SaveRect(cfg,"/Video/DialogRect",this->dlg_rect);
	cfg->Write("/Video/VideoEffect",this->video_effect);
}
	
//////////////////////////////////////////////////////////////////////////
//

mbMiscConfig::mbMiscConfig():
limit_speed(false),
show_status_bar(true),
show_menu(true),
fast_forward_disc_access(false),
model("b"),
window_rect(-1,-1,-1,-1),
window_is_maximized(false)
{
}

void mbMiscConfig::Load(mbWxConfig *cfg) {
	cfg->Read("/Misc/LimitSpeed",&this->limit_speed);
	cfg->Read("/Misc/DiscInterface",&this->disc_interface);
	cfg->Read("/Misc/ShowStatusBar",&this->show_status_bar);
	cfg->Read("/Misc/ShowMenu",&this->show_menu);
	cfg->Read("/Misc/FastForwardDiscAccess",&this->fast_forward_disc_access);
	cfg->Read("/Misc/Model",&this->model);
	LoadRect(cfg,"/Misc/WindowRect",&this->window_rect);
	cfg->Read("/Misc/WindowMaximized",&this->window_is_maximized);
	LoadRect(cfg,"/Misc/HardwareDialogRect",&this->hardware_dlg_rect);
}

void mbMiscConfig::Save(mbWxConfig *cfg) const {
	cfg->Write("/Misc/LimitSpeed",this->limit_speed);
	cfg->Write("/Misc/DiscInterface",this->disc_interface);
	cfg->Write("/Misc/ShowStatusBar",this->show_status_bar);
	cfg->Write("/Misc/ShowMenu",this->show_menu);
	cfg->Write("/Misc/FastForwardDiscAccess",this->fast_forward_disc_access);
	cfg->Write("/Misc/Model",this->model);
	SaveRect(cfg,"/Misc/WindowRect",this->window_rect);
	cfg->Write("/Misc/WindowMaximized",this->window_is_maximized);
	SaveRect(cfg,"/Misc/HardwareDialogRect",this->hardware_dlg_rect);
}

//////////////////////////////////////////////////////////////////////////
//
mbQuickstartConfig::mbQuickstartConfig():
dialog_pos(-1,-1,-1,-1)
{
}

void mbQuickstartConfig::Load(mbWxConfig *cfg) {
	unsigned i=0;
	bool done=false;
	do {
		wxString key,value;
		key.Printf("%u",i++);
		if(!cfg->Read("/Quickstart/Dirs/"+key,&value)) {
			done=true;
		} else {
			this->dirs.push_back(wxFileName::DirName(value));
			if(!this->dirs.back().IsOk()) {
				wxLogWarning("Ignoring invalid quickstart dir\n%s",value);
				this->dirs.pop_back();
			}
		}
	} while(!done);

	LoadRect(cfg,"/Quickstart/DialogRect",&this->dialog_pos);
}

void mbQuickstartConfig::Save(mbWxConfig *cfg)  const{
	cfg->DeleteGroup("/Quickstart/Dirs");
	for(unsigned i=0;i<this->dirs.size();++i) {
		wxString key;
		key.Printf("%u",i);
		cfg->Write("/Quickstart/Dirs/"+key,this->dirs[i].GetFullPath());
	}

	SaveRect(cfg,"/Quickstart/DialogRect",this->dialog_pos);
}

//////////////////////////////////////////////////////////////////////////
//

float HostInp_JoyState::*const mbJoysticksConfig::hostinp_axes[6]={
	&HostInp_JoyState::x,
	&HostInp_JoyState::y,
	&HostInp_JoyState::z,
	&HostInp_JoyState::rx,
	&HostInp_JoyState::ry,
	&HostInp_JoyState::rz,
};

mbJoysticksConfig::BeebStick::BeebStick():
enabled(false),
x_axis(X),
y_axis(Y),
button(-1)
{
}

mbJoysticksConfig::mbJoysticksConfig() {
}

void mbJoysticksConfig::Load(mbWxConfig *cfg) {
	for(unsigned i=0;i<2;++i) {
		wxString section=wxString::Format("/Joysticks/%u/",i);
		BeebStick *dest=&this->joysticks[i];
		int x_axis,y_axis;

		cfg->Read(section+"Enabled",&dest->enabled,false);
		cfg->Read(section+"Device",&dest->device_name,"");
		cfg->Read(section+"XAxis",&x_axis,INVALID_AXIS);
		cfg->Read(section+"YAxis",&y_axis,INVALID_AXIS);
		cfg->Read(section+"ButtonIndex",&dest->button,0);
		
		ClampValue(section+"ButtonIndex",&dest->button,-1,HostInp_num_joystick_buttons-1);
		ClampValue(section+"XAxis",&x_axis,0,int(INVALID_AXIS));
		ClampValue(section+"YAxis",&y_axis,0,int(INVALID_AXIS));

		dest->x_axis=Axis(x_axis);
		dest->y_axis=Axis(y_axis);
	}
}

void mbJoysticksConfig::Save(mbWxConfig *cfg) const {
	for(unsigned i=0;i<2;++i) {
		wxString section=wxString::Format("/Joysticks/%u/",i);
		const BeebStick *src=&this->joysticks[i];

		cfg->Write(section+"Enabled",src->enabled);
		cfg->Write(section+"Device",src->device_name);
		cfg->Write(section+"XAxis",src->x_axis);
		cfg->Write(section+"YAxis",src->y_axis);
		cfg->Write(section+"ButtonIndex",src->button);
	}
}

//////////////////////////////////////////////////////////////////////////
//
mbCmosConfig::mbCmosConfig() {
	for(unsigned i=0;i<bbcComputer::cmos_size;++i) {
		this->contents[i]=0;
	}
}

static t65::byte unhex(char ch) {
	if(isdigit(ch)) {
		return ch-'0';
	} else {
		return 10+tolower(ch)-'a';
	}
}

void mbCmosConfig::Load(mbWxConfig *cfg) {
	wxString value;

	cfg->Read("/Cmos/Contents",&value,"");

	if(value.find_first_not_of("0123456789ABCDEFabcdef")==wxString::npos&&
		value.length()==bbcComputer::cmos_size*2)
	{
		for(unsigned i=0;i<bbcComputer::cmos_size;++i) {
			this->contents[i]=unhex(value[i*2+0])*16+unhex(value[i*2+1]);
		}
	} else {
		memset(this->contents,0,sizeof this->contents);
	}
}

void mbCmosConfig::Save(mbWxConfig *cfg) const {
	wxString value;
	for(unsigned i=0;i<bbcComputer::cmos_size;++i) {
		value+=wxString::Format("%02X",this->contents[i]);
	}
	cfg->Write("/Cmos/Contents",value);
}

//////////////////////////////////////////////////////////////////////////

//static const char *default_filename="./beeb.ini";

static mbWxConfig *NewMbWxConfig(const char *filename) {
	long flags=wxCONFIG_USE_LOCAL_FILE|wxCONFIG_USE_RELATIVE_PATH;
	return new mbWxConfig("Beeb","Tom",filename,"",flags);
}

void mbConfig::GetConfig(mbWxConfig *cfg) const {
	this->roms.Save(cfg);
	this->sound.Save(cfg);
	this->keyboard.Save(cfg);
	this->video.Save(cfg);
	this->misc.Save(cfg);
	this->quickstart.Save(cfg);
	this->joysticks.Save(cfg);
	this->cmos.Save(cfg);
}

void mbConfig::Load(const wxString &config_file_name) {
	//This is to get around the possibility that a file specified without a
	//path will end up coming from some kind of systemwide config type place,
	//as is the case on Windows.
	mbWxConfig *conf=NewMbWxConfig(mbNormalizedFileName(config_file_name));//new mbWxConfig("mb");
	conf->SetExpandEnvVars(true);
	this->roms.Load(conf);
	this->sound.Load(conf);
	this->keyboard.Load(conf);
	this->video.Load(conf);
	this->misc.Load(conf);
	this->quickstart.Load(conf);
	this->joysticks.Load(conf);
	this->cmos.Load(conf);
	delete conf;
}

void mbConfig::SetFilename(const wxString &config_file_name,bool is_read_only) {
	this->read_only=is_read_only;
	this->file_name=wxFileName::FileName(config_file_name).GetFullPath();
}

void mbConfig::Save() const {
	if(this->read_only) {
		return;
	}
	mbWxConfig *conf=NewMbWxConfig(this->file_name);//new mbWxConfig("mb");
	this->GetConfig(conf);
	//Retry 10 times. This is to ensure Norton Anti Virus
	//doesn't bollocks it all up!
	//This is slightly silly if not on Windows, but hey! it won't do any harm.
	//TODO porting issue?
	int i;
	for(i=0;i<10;++i) {
		if(conf->Flush()) {
			break;
		}
#ifdef __WXMSW__
		Sleep(10);
#endif
	}
	delete conf;
	if(i==0) {
		//That's ok!
	} else if(i<10) {
		wxLogMessage("File saved OK despite warnings\nProbably Norton Anti Virus or Developer Studio...");
	} else {
		wxString tmp_name=wxFileName::CreateTempFileName(".");
		wxLogWarning("Saving cfg to tmp file \"%s\"\n",tmp_name.c_str());
		if(tmp_name) {
			conf=NewMbWxConfig(tmp_name);
			this->GetConfig(conf);
			if(!conf->Flush()) {
				wxLogWarning("Failed to save config to tmp file.\n");
			} else {
				wxLogWarning("Saved config successfully to tmp file.\n");
			}
			delete conf;
		}
		wxLogWarning("Failed to save config file. Check details.");
	}
	wxLog::FlushActive();
}

mbConfig::mbConfig():
read_only(false)
{
}
